# -*- coding: utf-8 -*-
"""
Created on Sat May 05 08:29:35 2012

@author: Owner
"""

#import python modules
import numpy as np
import itertools

#import compiled fortran libraries
import finallib

#define functions::

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


def LineSearch(Pos, Dir, dx, EFracTol, Type, M, L, Cut, Lambda,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""

    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [finallib.calcenergy(Pos + NormDir * x, Type, M, L, Cut, Lambda) 
                for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = finallib.calcenergy(Pos + NormDir * dx, Type, M, 
                        L, Cut, Lambda)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [finallib.calcenergy(Pos + NormDir * 2. * dx, Type, M, 
                        L, Cut, Lambda)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [finallib.calcenergy(Pos + NormDir * Dists[-1], Type, M, 
                        L, Cut, Lambda)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = finallib.calcenergy(Pos + NormDir * d3, Type, M, L, Cut, Lambda)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = finallib.calcenergy(Pos + NormDir * d3, Type, M, 
                        L, Cut, Lambda)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin,AtomTypes = Type,
                           AtomColorDict = {1:(1,0,0), 2:(0,0,1)})
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos,dx,Type,M,L,Cut,Lambda,EFracTolLS,EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""

    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = finallib.calcenergyforces(Pos,Type,M,L,Cut,Lambda,Forces)
    Dir = Forces
    OldPE = 1.e300
    i = 0;
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos,Dir,dx,EFracTolLS,Type,M,L,Cut,Lambda)
        OldForces = Forces.copy()
        PE, Forces = finallib.calcenergyforces(Pos,Type,M,L,Cut,Lambda,Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma * Dir
        if (i % 100 == 0):
            print i,PE,abs(PE - OldPE)/abs(PE)
        i += 1
    print i,PE,abs(PE - OldPE)/abs(PE)
    return PE, Pos

    
def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos,Type,M,L,Cut,Lambda):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""

    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = finallib.calcenergyforces(Pos,Type,M,L,Cut,Lambda,Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))
    
def PowerLaw(x, a, b):
    """Defines a simple power-law function for fitting curves.
Input: 
    x: (N,1) array of the independent variable
    a,b: fitting parameters
Output:
    y: power-law curve
"""
    return a*x**(b)
    
def SetTypes(N,A,B):
    """Defines an array containing the list of atom types for diblock 
copolymers.
Input:
    N: number of atoms in the simulation
    A: number of A atoms in a single polymer
    B: number of B atoms in a single polymer
Output:
    Type: (N,1) array of integers 1 or 2 corresponding to atom type
"""
    Type = np.zeros(N,'int')
    Names = range(N)
    flag = 1
    Name = 'H'
    j = 0
    for i in range(N):
        Type[i] = flag
        Names[i] = Name
        j += 1
        if j % A == 0:
            flag = 2
            Name = 'D'
        if j % (A+B) == 0:
            flag = 1
            Name = 'H'
            j = 0
    return (Type, Names)
    
def MinImage(Pos, L, PrintImages):
    """Returns a new Pos array with minimum-imaged positions."""
    center = np.mean(Pos,axis=0)
    Pos = Pos - center
    Pos = Pos - L*np.round_(Pos / L)
    if PrintImages:
        temp = np.copy(Pos)
        x = np.zeros_like(Pos); x[:,0] = L
        y = np.zeros_like(Pos); y[:,1] = L
        z = np.zeros_like(Pos); z[:,2] = L
        imagesx = [0,1]
        imagesy = [0,1]
        imagesz = [0,1]#[-1,0,1]
        for (xi,yi,zi) in itertools.product(imagesx,imagesy,imagesz):
            if abs(xi)+abs(yi)+abs(zi) > 0:
                Pos = np.append(Pos,temp + xi*x + yi*y +zi*z,axis=0)
    
    return Pos